#!/usr/bin/env nextflow
/*
main.nf
*
* --------------------------------------------------------------------------
* Licensed under MIT (https://git.biohpc.swmed.edu/BICF/Astrocyte/cellranger_count/blob/develop/LICENSE)
* --------------------------------------------------------------------------
*
*/

//  ########  ####  ######  ######## 
//  ##     ##  ##  ##    ## ##       
//  ##     ##  ##  ##       ##       
//  ########   ##  ##       ######   
//  ##     ##  ##  ##       ##       
//  ##     ##  ##  ##    ## ##       
//  ########  ####  ######  ##       


// Define Input variables
params.name = "run"
params.fastq = "test_data/mu.v3s1r500/*.fastq.gz"
params.designFile = "test_data/mu.v3s1r500/design.csv"
params.genome = 'mm10-2020-A'
params.expectCells = 10000
params.forceCells = 0
params.kitVersion = '3GEXv3'
params.version = '4.0.0'
params.vizFiles = true
params.astrocyte = false
params.outDir = "${baseDir}/output"

// Variable error test
if (params.kitVersion == "3GEXv3" && params.version == '2.1.1') {
  print("Cellranger Version 2.1.1 requires kitVersion 2")
  System.exit(1)	
}

// Define variables if astrocyte (or from config)
if (params.astrocyte) {
  print("Running under astrocyte")
  if (params.version == "4.0.0") {
    params.genomeLocation = '/project/apps_database/cellranger/refdata-gex-'
  } else {
    params.genomeLocation = '/project/apps_database/cellranger/refdata-cellranger-'
  }
  if (params.kitVersion == "3GEX") {
    params.chemistryParam ='threeprime'
  } else if (params.kitVersion == "3GEXv1") {
    params.chemistryParam ='SC3Pv1'
  } else if (params.kitVersion == "3GEXv2") {
    params.chemistryParam ='SC3Pv2'
  } else if (params.kitVersion == "3GEXv3") {
    params.chemistryParam ='SC3Pv3'
  } else if (params.kitVersion == "5GEX") {
    params.chemistryParam ='fiveprime'
  } else if (params.kitVersion == "5GEXPE") {
    params.chemistryParam ='SC5P-PE'
  } else if (params.kitVersion == "5GEXSE") {
    params.chemistryParam ='SC5P-R2'
  } else {
    params.chemistryParam = 'auto'
  }
} else {
  params.genomes = []
  params.genomeLocation = params.genome ? params.genomes[ params.genome ].loc ?: false : false
  params.chemistry = []
  params.chemistryParam = params.kitVersion ? params.chemistry[ params.kitVersion ].param ?: false : false
}
params.genomeLocationFull = params.genomeLocation+params.genome

// Define variables from input
pipelineVersion = "3.0.0_indev"
name = params.name
designLocation = Channel
  .fromPath(params.designFile)
  .ifEmpty { exit 1, "design file not found: ${params.designFile}" }
fastqList = Channel
  .fromPath(params.fastq)
  .flatten()
  .map { file -> [ file.getFileName().toString(), file.toString() ].join("\t") }
  .collectFile(name: 'fileList.tsv', newLine: true)
refLocation = Channel
  .fromPath(params.genomeLocationFull)
  .ifEmpty { exit 1, "referene not found: ${params.genome}" }
expectCells = params.expectCells
forceCells = params.forceCells
chemistryParam = params.chemistryParam
version = params.version
vizFiles = params.vizFiles
outDir = params.outDir

// Define script files
check_designScript = Channel.fromPath("$baseDir/scripts/check_design.py")
filename_checkScript = Channel.fromPath("$baseDir/scripts/filename_check.sh")
generate_versionsScript = Channel.fromPath("$baseDir/scripts/generate_versions.py")
generate_referencesScript = Channel.fromPath("$baseDir/scripts/generate_references.py")
versions_pythonScript = Channel.fromPath("$baseDir/scripts/versions_python.sh")
versions_seuratScript = Channel.fromPath("$baseDir/scripts/versions_seurat.sh")
downstream_vizScript = Channel.fromPath("$baseDir/scripts/downstream_viz.r")

// Define report files
multiqcConf = "${baseDir}/configs/multiqc_config.yaml"
references = "${baseDir}/../docs/references.md"

/*
 * trackStart: track start of pipeline
 */
params.ci = false
params.dev = false
process trackStart {
  script:
  """
  hostname
  ulimit -a
  export https_proxy=\${http_proxy}
 
  curl -H 'Content-Type: application/json' -X PUT -d '{ \
      "sessionId": "${workflow.sessionId}", \
      "pipeline": "cellranger_count", \
      "pipelineVersion": "${pipelineVersion}", \
      "start": "${workflow.start}", \
      "astrocyte": ${params.astrocyte}, \
      "status": "started", \
      "nextflowVersion": "${workflow.nextflow.version}",
      "ci": ${params.ci},
      "dev": ${params.dev}}' \
  "https://xku43pcwnf.execute-api.us-east-1.amazonaws.com/ProdDeploy/pipeline-tracking"
  """
}

/*
 * checkDesignFile: check design file for errors
 */
process checkDesignFile {
  tag "${name}"
  module 'python/3.6.1-2-anaconda'

  input:
    file designLocation
    file fastqList
    file check_designScript

  output:
    file("design.checked.csv") into designPaths

  script:
    """
    hostname
    ulimit -a
    noSpaceDesign=\$(echo "${designLocation}" | tr -d ' ')
    if [[ "\${noSpaceDesign}" != "${designLocation}" ]]; then
      mv "${designLocation}" "\${noSpaceDesign}"
    fi
    python3 check_design.py -d \${noSpaceDesign} -f ${fastqList}
    """
}


// Parse design file
samples = designPaths
  .splitCsv (sep: ',', header: true)
  .map { row -> [ row.Sample, file(row.fastq_R1), file(row.fastq_R2) ] }
  .groupTuple()
  .combine(filename_checkScript)
  //.subscribe { println it }

// Duplicate variables

samples.into {
  samples211
  samples302
  samples310
  samples400
}
refLocation.into {
  refLocation211
  refLocation302
  refLocation310
  refLocation400
}
expectCells211 = expectCells
expectCells302 = expectCells
expectCells310 = expectCells
expectCells400 = expectCells
forceCells211 = forceCells
forceCells302 = forceCells
forceCells310 = forceCells
forceCells400 = forceCells
chemistryParam211 = chemistryParam
chemistryParam302 = chemistryParam
chemistryParam310 = chemistryParam
chemistryParam400 = chemistryParam


/*
 * count211: run cellranger count version 2.1.1
 */
process count211 {
  tag "${sample}"
  publishDir "${outDir}/${task.process}", mode: 'copy'
  queue '128GB,256GB,256GBv1,384GB'
  module 'cellranger/2.1.1'

  input:
    set sample, file("${sample}_S?_L001_R1_001.fastq.gz"), file("${sample}_S?_L001_R2_001.fastq.gz"), file(script) from samples211
    file ref from refLocation211.first()
    expectCells211
    forceCells211
    chemistryParam211

  output:
    set sample, file("**/outs/**") into outPaths211
    set sample, file("**/outs/filtered_*/**"), file("**/outs/analysis/clustering/graphclust/**"), file("**/outs/analysis/clustering/kmeans_2_clusters/**"), file("**/outs/analysis/clustering/kmeans_3_clusters/**"), file("**/outs/analysis/clustering/kmeans_4_clusters/**"), file("**/outs/analysis/clustering/kmeans_5_clusters/**"), file("**/outs/analysis/clustering/kmeans_6_clusters/**"), file("**/outs/analysis/clustering/kmeans_7_clusters/**"), file("**/outs/analysis/clustering/kmeans_8_clusters/**"), file("**/outs/analysis/clustering/kmeans_9_clusters/**"), file("**/outs/analysis/clustering/kmeans_10_clusters/**"), file("**/outs/analysis/pca/**"), file("**/outs/analysis/tsne/**") into filteredOut211
    file("*_metrics_summary.tsv") into metricsSummary211

  when:
  version == '2.1.1'

  script:
    if (forceCells211 == 0) {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --expect-cells=${expectCells211} --chemistry=${chemistryParam211}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    } 
    else {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --force-cells=${forceCells211} --chemistry=${chemistryParam211}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
}


/*
 * count302: run cellranger count version 3.0.2
 */
process count302 {
  tag "${sample}"
  publishDir "${outDir}/${task.process}", mode: 'copy'
  queue '128GB,256GB,256GBv1,384GB'
  module 'cellranger/3.0.2'

  input:
    set sample, file("${sample}_S?_L001_R1_001.fastq.gz"), file("${sample}_S?_L001_R2_001.fastq.gz"), file(script) from samples302
    file ref from refLocation302.first()
    expectCells302
    forceCells302
    chemistryParam302

  output:
    set sample, file("**/outs/**") into outPaths302
    set sample, file("**/outs/filtered_*/**"), file("**/outs/analysis/clustering/graphclust/**"), file("**/outs/analysis/clustering/kmeans_2_clusters/**"), file("**/outs/analysis/clustering/kmeans_3_clusters/**"), file("**/outs/analysis/clustering/kmeans_4_clusters/**"), file("**/outs/analysis/clustering/kmeans_5_clusters/**"), file("**/outs/analysis/clustering/kmeans_6_clusters/**"), file("**/outs/analysis/clustering/kmeans_7_clusters/**"), file("**/outs/analysis/clustering/kmeans_8_clusters/**"), file("**/outs/analysis/clustering/kmeans_9_clusters/**"), file("**/outs/analysis/clustering/kmeans_10_clusters/**"), file("**/outs/analysis/pca/**"), file("**/outs/analysis/tsne/**") into filteredOut302
    file("*_metrics_summary.tsv") into metricsSummary302

  when:
    version == '3.0.2'

  script:
    if (forceCells302 == 0) {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --expect-cells=${expectCells302} --chemistry=${chemistryParam302}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    } 
    else {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --force-cells=${forceCells302} --chemistry=${chemistryParam302}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
}

/*
 * count310: run cellranger count version 3.1.0
 */
process count310 {
  tag "${sample}"
  publishDir "${outDir}/${task.process}", mode: 'copy'
  queue '128GB,256GB,256GBv1,384GB'
  module 'cellranger/3.1.0'

  input:
    set sample, file("${sample}_S?_L001_R1_001.fastq.gz"), file("${sample}_S?_L001_R2_001.fastq.gz"), file(script) from samples310
    file ref from refLocation310.first()
    expectCells310
    forceCells310
    chemistryParam310

  output:
    set sample, file("**/outs/**") into outPaths310
    set sample, file("**/outs/filtered_*/**"), file("**/outs/analysis/clustering/graphclust/**"), file("**/outs/analysis/clustering/kmeans_2_clusters/**"), file("**/outs/analysis/clustering/kmeans_3_clusters/**"), file("**/outs/analysis/clustering/kmeans_4_clusters/**"), file("**/outs/analysis/clustering/kmeans_5_clusters/**"), file("**/outs/analysis/clustering/kmeans_6_clusters/**"), file("**/outs/analysis/clustering/kmeans_7_clusters/**"), file("**/outs/analysis/clustering/kmeans_8_clusters/**"), file("**/outs/analysis/clustering/kmeans_9_clusters/**"), file("**/outs/analysis/clustering/kmeans_10_clusters/**"), file("**/outs/analysis/pca/**"), file("**/outs/analysis/tsne/**"), file("**/outs/analysis/umap/**") into filteredOut310
    file("*_metrics_summary.tsv") into metricsSummary310

  when:
    version == '3.1.0'

  script:
    if (forceCells310 == 0) {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --expect-cells=${expectCells310} --chemistry=${chemistryParam310}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
    else {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --force-cells=${forceCells310} --chemistry=${chemistryParam310}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
}

/*
 * count400: run cellranger count version 4.0.0
 */
process count400 {
  tag "${sample}"
  publishDir "${outDir}/${task.process}", mode: 'copy'
  queue '128GB,256GB,256GBv1,384GB'
  module 'cellranger/4.0.0'

  input:
    set sample, file("${sample}_S?_L001_R1_001.fastq.gz"), file("${sample}_S?_L001_R2_001.fastq.gz"), file(script) from samples400
    file ref from refLocation400.first()
    expectCells400
    forceCells400
    chemistryParam400

  output:
    set file("**/outs/**") into outPaths400
    set sample, file("**/outs/filtered_*/**"), file("**/outs/analysis/clustering/graphclust/**"), file("**/outs/analysis/clustering/kmeans_2_clusters/**"), file("**/outs/analysis/clustering/kmeans_3_clusters/**"), file("**/outs/analysis/clustering/kmeans_4_clusters/**"), file("**/outs/analysis/clustering/kmeans_5_clusters/**"), file("**/outs/analysis/clustering/kmeans_6_clusters/**"), file("**/outs/analysis/clustering/kmeans_7_clusters/**"), file("**/outs/analysis/clustering/kmeans_8_clusters/**"), file("**/outs/analysis/clustering/kmeans_9_clusters/**"), file("**/outs/analysis/clustering/kmeans_10_clusters/**"), file("**/outs/analysis/pca/**"), file("**/outs/analysis/tsne/**"), file("**/outs/analysis/umap/**") into filteredOut400
    file("*_metrics_summary.tsv") into metricsSummary400

  when:
    version == '4.0.0'

  script:
    if (forceCells400 == 0) {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --expect-cells=${expectCells310} --chemistry=${chemistryParam310}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
    else {
      """
      hostname
      ulimit -u 16384
      ulimit -a
      bash filename_check.sh -r ${ref}
      cellranger count --id=${sample} --transcriptome=./${ref} --fastqs=. --sample=${sample} --force-cells=${forceCells310} --chemistry=${chemistryParam310}
      sed -E 's/("([^"]*)")?(,|\$)/\\2\t/g' ${sample}/outs/metrics_summary.csv | tr -d "," | sed "s/^/${sample}\t/" > ${sample}_metrics_summary.tsv
      """
    }
}

// Collect all outputs reguardless of cellranger version
filteredOut = filteredOut211.mix(filteredOut302, filteredOut310, filteredOut400)
// Combine all inputs for downstreamViz
downstreamVizIn = downstream_vizScript.combine(versions_seuratScript).combine(filteredOut)

/*
 * downstreamViz: create files for downstream use (eg. R Seurat object)
 */
process downstreamViz {
  tag "${sample}"
  publishDir "${outDir}/seurat", mode: 'copy', pattern: "*.rds"
  module 'seurat/3.0.0'

  input:
    set file("*"), file("*"), sample, file("filtered/*"), file("clustering/graphclust/*"), file("clustering/kmeans_2_clusters/*"), file("clustering/kmeans_3_clusters/*"), file("clustering/kmeans_4_clusters/*"), file("clustering/kmeans_5_clusters/*"), file("clustering/kmeans_6_clusters/*"), file("clustering/kmeans_7_clusters/*"), file("clustering/kmeans_8_clusters/*"), file("clustering/kmeans_9_clusters/*"), file("clustering/kmeans_10_clusters/*"), file("pca/*"), file("tsne/*"), file("umap/*") from downstreamVizIn
    //file downstream_vizScript
    //file versions_seuratScript

  output:
    file "*.rds" into seuratPaths
    file "version_seurat.txt" into version_seurat

  when:
    vizFiles

  script:
    """
    hostname
    ulimit -a
    seurat-Rscript downstream_viz.r --sample ${sample} --cellrangerVersion ${version}
    bash versions_seurat.sh > version_seurat.txt
    """
}

/*
 * versions: collect all versions into a single yml
 */
process versions {
  tag "${name}"
  module 'python/3.6.1-2-anaconda:pandoc/2.7'

  input:
    file versions_pythonScript
    file version_seurat
    file generate_versionsScript
    file generate_referencesScript

  output:
    file("*.yaml") into yamlPaths

  script:
    """
    hostname
    ulimit -a
    echo "${workflow.nextflow.version}" > version_nextflow.txt
    echo "${pipelineVersion}" > version_pipeline.txt
    echo "${version}" > version_cellranger.txt
    bash versions_python.sh > version_python.txt
    python3 generate_versions.py -f version_*.txt -o versions
    python3 generate_references.py -r "${references}" -o references
    """
}

// Collect all metrics summaries reguardless of cellranger version
metricsSummary = metricsSummary211.mix(metricsSummary302, metricsSummary310, metricsSummary400)

/*
 * multiqc: create multiqc report
 */
process multiqc {
  tag "${name}"
  publishDir "${outDir}/${task.process}/${name}", mode: 'copy'
  module 'multiqc/1.7'

  input:
    file ('*') from metricsSummary.collect()
    file yamlPaths

  output:
    file "multiqc_report.html" into mqcPaths

  script:
    """
    hostname
    ulimit -a
    awk 'FNR==1 && NR!=1{next;}{print}' *.tsv > metrics_summary_mqc.tsv
    sed -i '1s/^.*\tE/Sample\tE/' metrics_summary_mqc.tsv
    multiqc -c ${multiqcConf} .
    """
}
